/* $XConsortium: Table.c,v 1.5 94/04/17 20:12:58 converse Exp $ */

/*

Copyright (c) 1987  X Consortium

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of the X Consortium shall not be
used in advertising or otherwise to promote the sale, use or other dealings
in this Software without prior written authorization from the X Consortium.

*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include "ngrid.h"


/*FIXME*/
enum {
  CHANGED,
  LAST_SIGNAL
};

static gint ngrid_signals[LAST_SIGNAL] = { 0 };


static void gtk_ngrid_class_init    (GtkNGridClass *klass);
static void gtk_ngrid_init          (GtkNGrid      *ngrid);
static void gtk_ngrid_realize       (GtkWidget           *widget);
static void gtk_ngrid_size_allocate (GtkWidget           *widget,
				     GtkAllocation       *allocation);
static void gtk_ngrid_size_request  (GtkWidget      *widget,
				     GtkRequisition *requisition);
gint gtk_ngrid_expose               (GtkWidget      *widget,
				     GdkEventExpose *event);
static void gtk_ngrid_adjustment_changed       (GtkAdjustment    *adjustment,
						gpointer          data);
static void gtk_ngrid_adjustment_value_changed (GtkAdjustment    *adjustment,
						gpointer          data);

static gint gtk_ngrid_focus_in  (GtkWidget     *widget,
				 GdkEventFocus *event) ;
static gint gtk_ngrid_focus_out (GtkWidget     *widget,
				 GdkEventFocus *event) ;

/* static void gtk_ngrid_draw (GtkWidget    *widget, */
/* 			       GdkRectangle *area); */

#define GtkNgrid GtkNGrid

guint
gtk_ngrid_get_type ()
{
  static guint ngrid_type = 0;

  if (!ngrid_type)
    {
      GtkTypeInfo ngrid_info =
      {
	"GtkNGrid",
	sizeof (GtkNGrid),
	sizeof (GtkNGridClass),
	(GtkClassInitFunc) gtk_ngrid_class_init,
	(GtkObjectInitFunc) gtk_ngrid_init,
	(GtkArgFunc) NULL,
      };
      ngrid_type = gtk_type_unique (gtk_widget_get_type (), &ngrid_info);
    }
  return ngrid_type;
}

static void
gtk_ngrid_class_init (GtkNGridClass *class)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;

  object_class = (GtkObjectClass*) class;
  widget_class = (GtkWidgetClass*) class;

  widget_class->realize = gtk_ngrid_realize;
  widget_class->size_allocate = gtk_ngrid_size_allocate;
  widget_class->size_request = gtk_ngrid_size_request;
  widget_class->expose_event = gtk_ngrid_expose;
  widget_class->focus_in_event = gtk_ngrid_focus_in;
  widget_class->focus_out_event = gtk_ngrid_focus_out;

/*   widget_class->button_press_event = gtk_ngrid_button_press; */
  ngrid_signals[CHANGED] =
    gtk_signal_new ("changed",
                    GTK_RUN_FIRST | GTK_RUN_NO_RECURSE,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (GtkNGridClass, changed),
                    gtk_signal_default_marshaller,
                    GTK_TYPE_NONE, 0);

}

static void
gtk_ngrid_init (GtkNGrid *ngrid)
{
/*   GTK_WIDGET_SET_FLAGS (ngrid, GTK_BASIC); */
  GTK_WIDGET_SET_FLAGS (ngrid, GTK_CAN_FOCUS);
/*   GTK_WIDGET_UNSET_FLAGS (ngrid, GTK_HAS_FOCUS); */
  /* set all the stff in a ngrid */
  ngrid->hadjustment = NULL;
  ngrid->vadjustment = NULL;
}

GtkWidget*
gtk_ngrid_new(gint rows, 
	      gint cols)
{
  GtkNGrid *ngrid;
  GtkAdjustment *hadjustment;
  GtkAdjustment *vadjustment;

  ngrid = gtk_type_new (gtk_ngrid_get_type ());
  ngrid->top_row = 1;
  ngrid->top_col = 1;
  ngrid->max_row = rows;
  ngrid->max_col = cols;
  ngrid->grid_lines = 1;
  ngrid->visible_cursor = 1;
  
  /* initialize rows and cols ? */
  /* maybe gc's ? */

  hadjustment = (GtkAdjustment*) gtk_adjustment_new (1.0, 1.0, (gfloat)cols,  
						     1.0, 5.0, 5.0);
  vadjustment = (GtkAdjustment*) gtk_adjustment_new (1.0, 1.0, (gfloat)rows,  
						     1.0, 5.0, 5.0);
  gtk_ngrid_set_hadjustment (ngrid, hadjustment);
  gtk_ngrid_set_vadjustment (ngrid, vadjustment);

  return GTK_WIDGET (ngrid);
}

GtkAdjustment*
gtk_ngrid_get_vadjustment (GtkNgrid *ngrid)
{
  g_return_val_if_fail (ngrid != NULL, NULL);
  g_return_val_if_fail (GTK_IS_NGRID (ngrid), NULL);

  return ngrid->vadjustment;
}

GtkAdjustment*
gtk_ngrid_get_hadjustment (GtkNgrid *ngrid)
{
  g_return_val_if_fail (ngrid != NULL, NULL);
  g_return_val_if_fail (GTK_IS_NGRID (ngrid), NULL);

  return ngrid->hadjustment;
}

void
gtk_ngrid_set_hadjustment (GtkNgrid   *ngrid,
			   GtkAdjustment *adjustment)
{
  g_return_if_fail (ngrid != NULL);
  g_return_if_fail (GTK_IS_NGRID (ngrid));
  g_return_if_fail (adjustment != NULL);
  
  if (ngrid->hadjustment)
    {
      gtk_signal_disconnect_by_data (GTK_OBJECT (ngrid->hadjustment), 
				     (gpointer) ngrid);
      gtk_object_unref (GTK_OBJECT (ngrid->hadjustment));
    }

  ngrid->hadjustment = adjustment;
  gtk_object_ref (GTK_OBJECT (ngrid->hadjustment));

  gtk_signal_connect (GTK_OBJECT (adjustment), "changed",
                      (GtkSignalFunc) gtk_ngrid_adjustment_changed,
                      (gpointer) ngrid);
  gtk_signal_connect (GTK_OBJECT (adjustment), "value_changed",
                      (GtkSignalFunc) gtk_ngrid_adjustment_value_changed,
                      (gpointer) ngrid);

  gtk_ngrid_adjustment_changed (adjustment, (gpointer) ngrid);
}

void
gtk_ngrid_set_vadjustment (GtkNgrid   *ngrid,
			   GtkAdjustment *adjustment)
{
  g_return_if_fail (ngrid != NULL);
  g_return_if_fail (GTK_IS_NGRID (ngrid));
  g_return_if_fail (adjustment != NULL);

  if (ngrid->vadjustment)
    {
      gtk_signal_disconnect_by_data (GTK_OBJECT (ngrid->vadjustment), 
				     (gpointer) ngrid);
      gtk_object_unref (GTK_OBJECT (ngrid->vadjustment));
    }

  ngrid->vadjustment = adjustment;
  gtk_object_ref (GTK_OBJECT (ngrid->vadjustment));

  gtk_signal_connect (GTK_OBJECT (adjustment), "changed",
                      (GtkSignalFunc) gtk_ngrid_adjustment_changed,
                      (gpointer) ngrid);
  gtk_signal_connect (GTK_OBJECT (adjustment), "value_changed",
                      (GtkSignalFunc) gtk_ngrid_adjustment_value_changed,
                      (gpointer) ngrid);

  gtk_ngrid_adjustment_changed (adjustment, (gpointer) ngrid);
}

static void
gtk_ngrid_adjustment_changed (GtkAdjustment *adjustment,
			      gpointer       data)
{
  GtkNgrid *ngrid;
  
  g_return_if_fail (adjustment != NULL);
  g_return_if_fail (data != NULL);
  g_return_if_fail (GTK_IS_NGRID (data));
  
  ngrid = GTK_NGRID (data);
}

static void
gtk_ngrid_adjustment_value_changed (GtkAdjustment *adjustment,
				    gpointer       data)
{
  GtkNgrid *ngrid;
  
  g_return_if_fail (adjustment != NULL);
  g_return_if_fail (data != NULL);
  g_return_if_fail (GTK_IS_NGRID (data));

  ngrid = GTK_NGRID (data);

  ngrid->top_row = (gint) ngrid->vadjustment->value;
  ngrid->top_col = (gint) ngrid->hadjustment->value;
  gtk_ngrid_expose(GTK_WIDGET(ngrid), NULL);
}


static void 
gtk_ngrid_size_request (GtkWidget      *widget,
		       GtkRequisition *requisition)
{
  GtkNGrid *ngrid;
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_NGRID (widget));
  g_return_if_fail (requisition != NULL);
  ngrid = GTK_NGRID (widget);
  requisition->width = GTK_WIDGET (ngrid)->requisition.width;
  requisition->height =   GTK_WIDGET (ngrid)->requisition.height;
}


static void
gtk_ngrid_size_allocate (GtkWidget     *widget,
			GtkAllocation *allocation)
{
  GdkEventConfigure event;
  
  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_NGRID (widget));
  g_return_if_fail (allocation != NULL);
  
  widget->allocation = *allocation;
  
  if (GTK_WIDGET_REALIZED (widget))
    {
      gdk_window_move_resize (widget->window,
			      allocation->x, allocation->y,
			      allocation->width, allocation->height);
      
      event.type = GDK_CONFIGURE;
      event.window = widget->window;
      event.x = allocation->x;
      event.y = allocation->y;
      event.width = allocation->width;
      event.height = allocation->height;
      
      gtk_widget_event (widget, (GdkEvent*) &event);
    }
}


void
gtk_ngrid_size (GtkNGrid *ngrid,
		gint            width,
		gint            height)
{
  g_return_if_fail (ngrid != NULL);
  g_return_if_fail (GTK_IS_NGRID (ngrid));

  GTK_WIDGET (ngrid)->requisition.width = width;
  GTK_WIDGET (ngrid)->requisition.height = height;
}

static int inblock(GtkNGrid *tw, int r, int c)
{
  return ((r >= tw->sel_top_row) &&
	  (r <= tw->sel_bottom_row) &&
	  (c >= tw->sel_left_col) &&
	  (c <= tw->sel_right_col));
}

/* return the type of the cell */
/* if a query function isn't defined, assume TEXT */
static int ret_type(GtkNGrid *tw, int row, int col)
{
  if (tw->type)
    return (*tw->type)(tw, tw->data, row, col);
  return NGRID_TEXT;
}

/* return the printable value for the cell */
/* if a query function isn't defined, return NULL */
static char *ret_pvalue(GtkNGrid *tw, int row, int col)
{
  if (tw->text)
    return (*tw->text)(tw, tw->data, row, col);
  else return NULL;
}

static GtkNGridEntryFeatures *ret_features(GtkNGrid *tw, int row, int col)
{
  if (tw->features)
    return (*tw->features)(tw, tw->data, row, col);
  return NULL;
}

static gint cell_width(GtkNGrid *tw, int col)
{
  gint wi;
  
  if (tw->col_width) {
    wi = (*tw->col_width)(tw, tw->data, col);
    return wi;
  }
  return tw->default_width;
}

static gint cell_height(GtkNGrid *tw, int row)
{
  gint h;
  
  if (tw->row_height) {
    h = (*tw->row_height)(tw, tw->data, row);
    return h;
  }
  return tw->default_height;
}

static void draw_cell(GtkNGrid *tw, GdkPixmap *scribble,
			int row, int col, int x_base, int y_base)
{
  GtkWidget *w;
  gchar *p;
  GdkFont *font;                  
  GdkGC *gc;
  GtkNGridEntryFeatures *features;
  gint x_pos, y_pos;
  guint text_height, text_width;
  gulong format = 0;
  guchar highlight = 0;
  gint cw, ch;

  w = GTK_WIDGET(tw);
  if (row > tw->max_row || col > tw->max_col) return;

#ifdef 0  
  if (ret_type(tw, row, col) == NGRID_EMBED) {
    char *p = ret_pvalue(tw, row, col);
    embed_draw(scribble, x_base, y_base, p);  /* ? */
    return;
  }
#endif

  if (ret_type(tw, row, col) == NGRID_PIXMAP) {
    g_warning("The ngrid widget does not yet display pixmaps");
    return;
  }
  
  if ((features = ret_features(tw, row, col))) {
    format = features->format;
    gc = features->gc?features->gc:w->style->fg_gc[GTK_STATE_NORMAL];
    font = features->font?features->font:GTK_WIDGET(tw)->style->font;
    highlight = features->highlight;
  } else {
    gc = w->style->fg_gc[GTK_STATE_NORMAL];
    font = GTK_WIDGET(tw)->style->font;
  }    

  p = ret_pvalue(tw, row, col);
  if (p == NULL) p = "";
  
#if 1	/* this seems to be necessary to keep X from crashing */
  if (strlen(p) > 1000) p[1000] = '\0';
#endif

  text_height = font->ascent;
  text_width = gdk_string_width(font, p);
  cw = cell_width(tw, col);
  ch = cell_height(tw, row);

  if (cw<=0 || ch <=0) return;

  if (highlight) {
    GdkGCValues values;
    gdk_gc_get_values(gc,&values);
    gdk_gc_set_foreground(gc,&(values.background));
    gdk_draw_rectangle(scribble, gc,
		       TRUE,
		       x_base, y_base,
		       cw, ch);
    gdk_gc_set_foreground(gc,(&values.foreground));
  }

  /* FIXME */
  switch (format) {
  case GTK_JUSTIFY_CENTER:
    x_pos = (cw-text_width) / 2;
    break;
  case GTK_JUSTIFY_RIGHT:
    x_pos = cw-text_width-5;
    break;
  default:	/* GTK_JUSTIFY_LEFT */
    x_pos = 5;
  }
  
  switch (format) {
  case GTK_POS_BOTTOM:
    y_pos = ch-5;
    break;
  case GTK_POS_TOP:
    y_pos = text_height+font->descent;
    break;
  default:	/* GTK_JUSTIFY_CENTER */
    y_pos = (ch+text_height)/2
      -font->descent;
  }
  
  gdk_draw_string (scribble,
		   font,
		   gc,
		   x_base+x_pos, y_base+y_pos,
		   p);
  
  if (tw->grid_lines) {
    gdk_draw_line(scribble, GTK_WIDGET(tw)->style->fg_gc[GTK_STATE_NORMAL],
		  x_base, y_base+ch,
		  x_base+cw, y_base+ch);
    gdk_draw_line(scribble, GTK_WIDGET(tw)->style->fg_gc[GTK_STATE_NORMAL],
		  x_base+cw, y_base+ch,
		  x_base+cw, y_base);
  }
#ifdef 0
  /* these lines are from the format */
  if (format & BORDER_LEFT)
    gdk_draw_line(scribble, GTK_WIDGET(tw)->style->fg_gc[GTK_STATE_NORMAL],
		  x_base+1, y_base+1,
		  x_base+1, y_base+ch-1);
  if (format & BORDER_RIGHT)
    gdk_draw_line(scribble, GTK_WIDGET(tw)->style->fg_gc[GTK_STATE_NORMAL],
		  x_base+cw-1, y_base+1,
		  x_base+cw-1, y_base+ch-1);
  if (format & BORDER_TOP)
    gdk_draw_line(scribble, GTK_WIDGET(tw)->style->fg_gc[GTK_STATE_NORMAL],
		  x_base+1, y_base+1,
		  x_base+cw-1, y_base+1);
  if (format & BORDER_BOTTOM)
    gdk_draw_line(scribble, GTK_WIDGET(tw)->style->fg_gc[GTK_STATE_NORMAL],
		  x_base+1, y_base+ch-1,
		  x_base+cw-1, y_base+ch-1);
#endif /* 0 */
  if (0 && inblock(tw, row, col)) 
    gdk_draw_rectangle(scribble, tw->xor_gc,
		       TRUE,
		       x_base, y_base,
		       cw, ch);
  if (features && features->free_me) 
    g_free(features);
}

static void cell_row(GtkNGrid *tw, GdkPixmap *scribble,
		     int width, int y_base, int i)
{
  int j, x_base;
  x_base = 0;
  
  for (j = 0; (j < tw->prot_col) && (x_base < width); j++) {
    draw_cell(tw, scribble, i, j, x_base, y_base);
    x_base += cell_width(tw, j);
  }
  for (j = tw->top_col; x_base < width; j++) {
    draw_cell(tw, scribble, i, j, x_base, y_base);
    x_base += cell_width(tw, j);
  }
}

static void draw_ngrid(GtkNGrid *tw, GdkPixmap *d)
{
  gint y_base = 0;
  gint width = GTK_WIDGET(tw)->allocation.width;
  gint height = GTK_WIDGET(tw)->allocation.height;
  gint i;
  
  for (i = 0; (i < tw->prot_row) && (y_base < height); i++) {
    cell_row(tw, d, width, y_base, i);
    y_base += cell_height(tw, i);
  }
  for (i = tw->top_row; y_base < height; i++) {
    cell_row(tw, d, width, y_base, i);
    y_base += cell_height(tw, i);
  }
}

/* create, draw and return a pixmap of the stage at time now.
   reuse pixmap from previous call if it's the right size
   (this is more efficient than the old strategy: malloc & free each time
   and also slows the memory leak caused by bug in gdkpixmap.c 
   from a raging torrent 
   to a slow trickle 
   (now leaks memory only when window is resized)
*/
/* Strategy:
   1. Clear the window, or create a new empty Pixmap
   2. Paint a rectangle for the block, if there is one (coords != 0)
   3. Draw the grid
   4. Draw the data
   5. Draw the cursor using XOR, if one is desired (coords != 0)
   The purpose of using XOR is to make sure the cursor is visible
   on any background. Technicolor is a potential side effect.
*/

GdkPixmap *ngrid_pixmap(GtkNGrid *tw)
{
  static gint scribble_width = 0;
  static gint scribble_height = 0;
  static GdkPixmap *scribble = NULL;

  gint width = GTK_WIDGET(tw)->allocation.width;
  gint height = GTK_WIDGET(tw)->allocation.height;
  
  if (!tw->non_gr_exp_gc)
    {
      tw->non_gr_exp_gc = gdk_gc_new (GTK_WIDGET(tw)->window);
      gdk_gc_set_exposures (tw->non_gr_exp_gc, FALSE);
    }
  
  if (width!=scribble_width || height!=scribble_height) {
    (scribble != NULL) ?
      gdk_pixmap_destroy(scribble):  /* deprecated, but gdk_pixmap_unref() isn't implemented */
      "fnord";
    scribble = gdk_pixmap_new(GTK_WIDGET(tw)->window,
			      width, height, -1); 
    scribble_height = height;
    scribble_width = width;
  }

  gdk_draw_rectangle(scribble, GTK_WIDGET(tw)->style->white_gc,
		     TRUE,
		     0, 0,
		     width, height);

  draw_ngrid(tw, scribble);
  return scribble;
}

static gint
gtk_ngrid_focus_in (GtkWidget     *widget,
                    GdkEventFocus *event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_NGRID (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  GTK_WIDGET_SET_FLAGS (widget, GTK_HAS_FOCUS);
  gtk_widget_draw_focus (widget);

  return FALSE;
}

static gint
gtk_ngrid_focus_out (GtkWidget     *widget,
                     GdkEventFocus *event)
{
  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_NGRID (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  GTK_WIDGET_UNSET_FLAGS (widget, GTK_HAS_FOCUS);
  gtk_widget_draw_focus (widget);

  return FALSE;
}


static void
gtk_ngrid_realize (GtkWidget *widget)
{
  GtkNGrid *ngrid;
  GdkWindowAttr attributes;
  GdkGCValues values;
  gint attributes_mask;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_NGRID (widget));

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
  ngrid = GTK_NGRID (widget);

  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget);
  attributes.event_mask |= (GDK_EXPOSURE_MASK |
			    GDK_ENTER_NOTIFY_MASK |
			    GDK_LEAVE_NOTIFY_MASK |
			    GDK_KEY_PRESS_MASK |
			    GDK_KEY_RELEASE_MASK |
			    GDK_BUTTON_PRESS_MASK);

  /* FIXME: add all other wanted events */

  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

  widget->window = gdk_window_new (widget->parent->window, &attributes, 
				   attributes_mask);
  gdk_window_set_user_data (widget->window, ngrid);

  widget->style = gtk_style_attach (widget->style, widget->window);
  gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL);

  gdk_window_set_background (widget->window, &widget->style->white);

  ngrid->non_gr_exp_gc = gdk_gc_new (widget->window);
  gdk_gc_set_exposures (ngrid->non_gr_exp_gc, FALSE);

  values.foreground = widget->style->white;
  values.function = GDK_XOR;
  values.line_width = 2;
  values.line_style = GDK_LINE_SOLID;
  values.cap_style = GDK_CAP_BUTT;;
  values.join_style = GDK_JOIN_MITER;
  values.subwindow_mode = GDK_INCLUDE_INFERIORS;
  ngrid->xor_gc = gdk_gc_new_with_values (widget->window,
					  &values,
					  GDK_GC_LINE_WIDTH |
					  GDK_GC_LINE_STYLE |
					  GDK_GC_CAP_STYLE |
					  GDK_GC_FOREGROUND |
					  GDK_GC_JOIN_STYLE |
					  GDK_GC_FUNCTION |
					  GDK_GC_SUBWINDOW);


/*   gdk_window_set_background (grid->window, &widget->style->white); */

/*   gdk_window_show (grid->window); */
}

static gint cell_next_row(GtkNGrid *tw, int row)
{
        if (row+1 == tw->prot_row) return tw->top_row;
        return row+1;
}

static gint cell_next_col(GtkNGrid *tw, int col)
{
        if (col+1 == tw->prot_col) return tw->top_col;
        return col+1;
}

static gint cell_prev_row(GtkNGrid *tw, int row)
{
        if (row == tw->top_row) return tw->prot_row-1;
        return row-1;
}

static int cell_prev_col(GtkNGrid *tw, int col)
{
        if (col == tw->top_col) return tw->prot_col-1;
        return col-1;
}

void gtk_ngrid_rowcol_to_coord(GtkNGrid *ngrid,
			       gint cell_row, gint cell_col,
			       gint *cell_x,  gint *cell_y)
{
        int i;

        *cell_y = 0;
        for (i = 0; i < ngrid->prot_row; i++)
                *cell_y += cell_height(ngrid, i);

        while (cell_row < ngrid->top_row) {
                cell_row = cell_next_row(ngrid, cell_row);
                *cell_y -= cell_height(ngrid, cell_row);
        }
        while (cell_row > ngrid->top_row) {
                cell_row = cell_prev_row(ngrid, cell_row);
                *cell_y += cell_height(ngrid, cell_row);
        }
        *cell_x = 0;
        for (i = 0; i < ngrid->prot_col; i++)
                *cell_x += cell_width(ngrid, i);

        while (cell_col < ngrid->top_col) {
                cell_col = cell_next_col(ngrid, cell_col);
                *cell_x -= cell_width(ngrid, cell_col);
        }
        while (cell_col > ngrid->top_col) {
                cell_col = cell_prev_col(ngrid, cell_col);
                *cell_x += cell_width(ngrid, cell_col);
        }
}

void gtk_ngrid_coord_to_rowcol(GtkNGrid *tw,
			       int cur_x, int cur_y,
			       int *cur_row, int *cur_col)
{
  int prot_x = 0, prot_y = 0, i;
  
  for (i = 0; i < tw->prot_col; i++)
    /*              prot_x += cell_width(w->buf, i);
     */              cur_x -= cell_width(tw, i);
  for (i = 0; i < tw->prot_row; i++)
    /*              prot_y += cell_height(w->buf, i);
     */              cur_y -= cell_height(tw, i);
  
  *cur_row = tw->top_row;
  *cur_col = tw->top_col;
  while (cur_y < prot_y) {
    cur_y += cell_height(tw, *cur_row);
    (*cur_row) = cell_prev_row(tw, *cur_row);
  }
  while (cur_y > cell_height(tw, *cur_row)) {
    cur_y -= cell_height(tw, *cur_row);
    (*cur_row) = cell_next_row(tw, *cur_row);
  }
  while (cur_x < prot_x) {
    cur_x += cell_width(tw, *cur_col);
    (*cur_col) = cell_prev_col(tw, *cur_col);
  }
  while (cur_x > cell_width(tw, *cur_col)) {
    cur_x -= cell_width(tw, *cur_col);
    (*cur_col) = cell_next_col(tw, *cur_col);
  }
  
}

static void toggle_cursor(GtkNGrid *tw)
{
  int cur_x, cur_y, w, h, row, col;
  row = tw->point_row;
  col = tw->point_col;

  if (row < 0 || (row >tw->prot_row && row <= tw->top_row)) return;
  if (col < 0 || (col >tw->prot_col && col <= tw->top_col)) return;

  gtk_ngrid_rowcol_to_coord(tw, row, col,
			    &cur_x, &cur_y);

  w = cell_width(tw, tw->point_col);
  h = cell_height(tw, tw->point_row);
  if (w<4 || h<4) return;
  gdk_draw_rectangle(GTK_WIDGET(tw)->window,
		     tw->xor_gc, 
		     FALSE,
		     cur_x+3, cur_y+3,
		     w-5,
		     h-5);
}

/* FIXME: should be static */
gint
gtk_ngrid_expose (GtkWidget      *widget,
		  GdkEventExpose *event)
{
  GtkNGrid *ngrid;
  GdkPixmap *scribble;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_NGRID (widget), FALSE);

  if (GTK_WIDGET_VISIBLE (widget) && GTK_WIDGET_MAPPED (widget))
    {
      ngrid = GTK_NGRID (widget);
      scribble = ngrid_pixmap(ngrid);
/*       gdk_window_clear_area (widget->window, */
/* 			     0, 0, */
/* 			     widget->allocation.width, */
/* 			     widget->allocation.height); */
      gdk_draw_pixmap(widget->window,
 		      ngrid->non_gr_exp_gc,
		      scribble, 0, 0, 0, 0,
		      widget->allocation.width-1,
		      widget->allocation.height-1);

      /* draw the cursor, perhaps? */
      if (ngrid->visible_cursor) {
	toggle_cursor(ngrid);
      }
    }
  return TRUE;
}

void gtk_ngrid_set_point (GtkNGrid *ngrid,
			  gint row, gint col) 
{
  ngrid->point_row = row;
  ngrid->point_col = col;
  gtk_ngrid_expose(GTK_WIDGET(ngrid),NULL);

}  

void gtk_ngrid_get_point (GtkNGrid *ngrid,
			  gint *row, gint *col) 
{
  *row = ngrid->point_row;
  *col = ngrid->point_col;
}  


#undef GtkNgrid
